import re


class ReplaceDict:
    """Replaces a series of keys to values."""

    def __init__(self, rep_dict: dict):
        """

        :param rep_dict: A dict, in which all keys and values must be str.
        """
        self.rep_dict = rep_dict
        self.pattern = re.compile(
            '|'.join(map(re.escape, self.rep_dict.keys())), re.M)

    def replace(self, string: str) -> str:
        """Replaces the given str according to the pattern."""
        return self.pattern.sub(lambda key: self.rep_dict[key.group()], string)



from ec_tools.basic_tools.replace_dict import ReplaceDict
from ec_tools import basic_tools


class ColorfulStr:
    def __init__(self):
        self.colors = {
            'black': 30,
            'red': 31,
            'green': 32,
            'yellow': 33,
            'blue': 34,
            'magenta': 35,
            'cyan': 36,
            'white': 37,
            'r': 31,
            'g': 32,
            'y': 33,
            'b': 34,
        }
        self.color_format = '\033[1;{}m'
        self.default = '\033[1;0m'
        table = {
            '(#{})'.format(k): self.color_format.format(v)
            for k, v in self.colors.items()
        }
        table['(#)'] = self.default
        self.replace_colors = ReplaceDict(table)
        self.clean_colors = ReplaceDict({k: '' for k in table.keys()})
        self.done = self('(#g)done(#)')

    def __call__(self, *args) -> str:
        output = basic_tools.touch_suffix(
            ' '.join(map('{}'.format, args)), '(#)')
        output = self.replace_colors.replace(output)
        return output

    def color(self, *args, color='(#g)') -> str:
        return self.replace_colors.replace(color) + self(*args)

    def green(self, *args) -> str:
        return self.color(*args, color='(#g)')

    def red(self, *args) -> str:
        return self.color(*args, color='(#r)')

    def yellow(self, *args) -> str:
        return self.color(*args, color='(#y)')

    def blue(self, *args) -> str:
        return self.color(*args, color='(#b)')

    def clean(self, *args) -> str:
        output = ' '.join(map('{}'.format, args))
        return self.clean_colors.replace(output)


colorful_str = ColorfulStr()



import os
import time


def touch_prefix(target: str, prefix: str) -> str:
    return target if target.startswith(prefix) else (prefix + target)


def touch_suffix(target: str, suffix: str) -> str:
    return target if target.endswith(suffix) else (target + suffix)


def mkdir(path: str, dirname_check=True):
    dir_path = os.path.dirname(path) if dirname_check else path
    if not os.path.isdir(dir_path):
        os.makedirs(dir_path)


def get_cur_time(time_format='%Y-%m-%d %H:%M:%S'):
    return time.strftime(time_format, time.localtime())



def get_headers(path):
    headers = {}
    with open(path) as f:
        for row in f.readlines():
            row = row.strip()
            index = row.find(":")
            headers[row[:index]] = row[index + 1:].strip()
    return headers



from ec_tools.basic_tools.colorful_log import ColorfulLog
from ec_tools.basic_tools.colorful_str import colorful_str
from ec_tools.basic_tools.procedure import Procedure as procedure
from ec_tools.basic_tools.replace_dict import ReplaceDict

from ec_tools import basic_tools
from ec_tools import database
from ec_tools import image
from ec_tools import spider



import logging
import os
import sys
from traceback import TracebackException

from ec_tools.basic_tools.colorful_str import colorful_str
from ec_tools import basic_tools

DEFAULT_DETAILED_LOG_FORMAT = '[(#y)%(levelname)s(#) (#b)%(filename)s/%(module)s/%(funcName)s/#L%(lineno)d(#)' \
                              ' (#g)%(asctime)s(#)] %(message)s'
DEFAULT_SIMPLE_LOG_FORMAT = '(#y)[%(levelname)s](#) %(message)s'


def create_stream_handle(level, formatter: str):
    stream_handle = logging.StreamHandler()
    stream_handle.setLevel(level)
    stream_handle.setFormatter(
        logging.Formatter(colorful_str(formatter)))
    return stream_handle


def create_file_handle(path: str, level: int, formatter: str):
    file_handle = logging.FileHandler(path)
    file_handle.setLevel(level)
    file_handle.setFormatter(
        logging.Formatter(colorful_str.clean(formatter)))
    return file_handle


class ColorfulLog(logging.Logger):
    def __init__(
            self,
            log_level=logging.INFO,
            log_formatter=DEFAULT_SIMPLE_LOG_FORMAT,
            log_dir='logs',
            log_name='colorful_log',
    ):
        f"""

        :param log_level: one of {logging.DEBUG, logging.INFO, logging.WARN, logging.ERROR, logging.CRITICAL}
        :param log_formatter: log format, default DEFAULT_SIMPLE_LOG_FORMAT, candidate DEFAULT_DETAILED_LOG_FORMAT
        :param log_dir: default "logs/", None for no file log
        :param log_name: name of logger
        """
        super().__init__(log_name, log_level)

        if log_dir is None:
            self.log_path = None
        else:
            self.log_path = os.path.join(
                log_dir, basic_tools.touch_suffix(log_name, '.log'))
            basic_tools.mkdir(self.log_path)
            self.addHandler(
                create_file_handle(path=self.log_path,
                                   level=log_level,
                                   formatter=log_formatter))
        self.addHandler(
            create_stream_handle(level=log_level, formatter=log_formatter))

    def print_exception(self, limit=None, chain=True):
        etype, value, tb = sys.exc_info()
        msg = '\n'.join(TracebackException(type(value), value, tb, limit=limit).format(chain=chain))
        self.error(msg)
        self.error(etype)


ec_tools_local_logger = ColorfulLog(log_dir=None, log_name='ec_tools')



import itertools
import re

from ec_tools.database.database_client import DatabaseClientInterface

PARENTHESES_PATTERN = re.compile(r'(\([^()]+\))')


def get_col_num(sql_pattern):
    return PARENTHESES_PATTERN.findall(sql_pattern)[0].count(',') + 1


def batch_execute(database_client: DatabaseClientInterface, sql_pattern: str, items: list):
    if not items:
        return None
    if not isinstance(items[0], (list, tuple)):
        items = [items]
    col_num = get_col_num(sql_pattern)
    assert col_num == len(items[0]), 'size of sql and data does not match! {} != {}'.format(
        col_num, len(items[0]))
    sql = sql_pattern.format(', '.join(['({})'.format(', '.join(['?'] * col_num))] * len(items)))
    args = list(itertools.chain(*items))
    return database_client.execute(sql, args)



import abc


class DatabaseClientInterface(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def commit(self):
        pass



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

packages = {
    'ec_tools': 'src/ec_tools',
    'ec_tools.basic_tools': 'src/ec_tools/basic_tools',
    'ec_tools.database': 'src/ec_tools/database',
    'ec_tools.spider': 'src/ec_tools/spider',
    'ec_tools.image': 'src/ec_tools/image',
}

setuptools.setup(
    name='ec_tools',
    version='0.8',
    packages=packages.keys(),
    package_dir=packages,
    url='http://github.com/indestinee/ec_tools',
    author='ec',
    author_email='indestinee@gmail.com',
    description='tools',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
)



import datetime
import time

import ec_tools
from ec_tools.basic_tools.colorful_log import ec_tools_local_logger


class Procedure:
    def __init__(self, msg: str, logger=ec_tools_local_logger):
        self.logger = logger
        self.logs = [msg]
        self.time = time.time()

    def __enter__(self):
        self.logger.info(*self.logs)
        return self

    def add_log(self, *args):
        self.logs += args

    def __exit__(self, type, value, traceback):
        log = self.logger.info if traceback is None else self.logger.error
        log(ec_tools.colorful_str(
            *self.logs,
            ec_tools.colorful_str.done,
            'cost time: (#b){}'.format(
                datetime.timedelta(seconds=time.time() - self.time))))



import numpy as np


def bytes_to_img(data: bytes) -> np.ndarray:
    import cv2
    np_arr = np.frombuffer(data, np.uint8)
    img = cv2.imdecode(np_arr, cv2.IMREAD_COLOR)
    return img


def random_image(size=(256, 256, 3)):
    return np.random.randint(256, size=size, dtype=np.uint8)



import sqlite3
import threading
from ec_tools import basic_tools
from ec_tools.database.database_client import DatabaseClientInterface
from ec_tools.basic_tools.colorful_log import ec_tools_local_logger

DICT_FACTORY = lambda cursor, row: {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


class SqliteClient(DatabaseClientInterface):
    def __init__(self, name, logger=ec_tools_local_logger, return_dict=False):
        super().__init__()
        self.db_name = basic_tools.touch_suffix(name, '.db')
        self.logger = logger
        self.return_dict = return_dict
        self.lock = threading.Lock()
        self.conn, self.cursor = None, None
        self.__connect_db()
        self.execute("PRAGMA FOREIGN_KEYS=ON")

    def commit(self):
        self.conn.commit()

    def execute(self, sqls, args=None):
        if args is None:
            args = []
        with self.lock:
            if isinstance(sqls, str):
                result = self.__execute_one(sqls, args)
                self.commit()
                return result
            else:
                num_q = sum([sql.count('?') for sql in sqls])
                assert num_q == len(args), 'num(?) != len(args): {} != {}'.format(num_q, len(args))
                start_index, results = 0, []
                for sql in sqls:
                    params_cnt = sql.count('?')
                    result = self.__execute_one(sql, args[start_index: start_index + params_cnt])
                    results.append(result)
                    start_index += params_cnt
                self.commit()

    def __execute_one(self, sql, args):
        try:
            result = self.cursor.execute(sql, args)
        except sqlite3.OperationalError as e:
            self.logger.warn(e)
            self.logger.print_exception()
            self.__connect_db()
            result = self.cursor.execute(sql, args)
        if isinstance(result, sqlite3.Cursor):
            result = result.fetchall()
        return result

    def __connect_db(self):
        self.conn = sqlite3.connect(self.db_name, check_same_thread=False)
        if self.return_dict:
            self.conn.row_factory = DICT_FACTORY
        self.cursor = self.conn.cursor()
